/* eslint-disable @typescript-eslint/ban-types */
/* eslint-disable no-param-reassign */

import axios, {
  AxiosResponse,
  AxiosRequestConfig,
  CancelTokenStatic,
  AxiosInstance,
} from 'axios';

import qs from 'qs';

import {
  API_URL,
  requestTimeout,
} from '../../config/net.config';

class MyRequest {
  // `API_URL` 将自动加在 `url` 前面，除非 `url` 是一个绝对 URL。
  protected service!: AxiosInstance;

  protected pending: Array<{ url: string, cancel: Function }> = [];

  protected CancelToken: CancelTokenStatic = axios.CancelToken;

  protected axiosRequestConfig: AxiosRequestConfig = {};

  private static resInstance: MyRequest;

  constructor() {
    this.requestConfig();
    this.service = axios.create(this.axiosRequestConfig);
    this.interceptorsRequest();
    this.interceptorsResponse();
  }

  /**
   * 初始化配置
   * @protected
   */
  protected requestConfig(): void {
    this.axiosRequestConfig = {
      baseURL: API_URL,
      timeout: requestTimeout,
    };
  }

  /**
   * 取消所有还未获取到数据的请求
   * @public
   */
  public cancelPreRequests() {
    this.pending.forEach((s) => {
      s.cancel('cancel from client');
    });
    this.pending = [];
  }

  /**
   * 获取请求配置拼装的key
   * @param config
   * @protected
   */
  // eslint-disable-next-line class-methods-use-this
  protected getKeyOfRequest(config: AxiosRequestConfig): string {
    let key = config.url;
    if (config.params) key += JSON.stringify(config.params);
    if (config.data) key += JSON.stringify(config.data);
    key += `&request_type=${config.method}`;
    return key as string;
  }

  /**
   * 请求拦截
   * @protected
   */
  protected interceptorsRequest(): void {
    this.service.interceptors.request.use(
      (config: AxiosRequestConfig) => {
        const keyOfRequest = this.getKeyOfRequest(config);
        config.cancelToken = new this.CancelToken((c) => {
          this.pending.push({
            url: keyOfRequest,
            cancel: c,
          });
        });
        return config;
      },
      (error) => Promise.reject(error),
    );
  }

  /**
   * 响应拦截
   * @protected
   */
  protected interceptorsResponse(): void {
    this.service.interceptors.response.use(
      (response: AxiosResponse) => this.handleResponse(response),
      (error) => {
        const { response } = error;
        if (response === undefined) {
          return Promise.reject(new Error(error));
        }
        return this.handleResponse(response);
      },
    );
  }

  /**
   * 请求特殊处理
   * @protected
   * @param response
   */
  // eslint-disable-next-line class-methods-use-this
  protected handleResponse<T>(response: AxiosResponse<T>): Promise<AxiosResponse<T>> {
    const code = response.status;
    switch (code) {
      case 200:
        return Promise.resolve(response);
      case 401:
        // TODO token失效,跳转登录页
        break;
      case 403:
        // TODO 没有权限,跳转403页面
        break;
      default:
        break;
    }
    return Promise.reject(response);
  }

  /**
   * post 方法
   * @param url
   * @param data
   * @param config
   */
  public post<T>(
    url: string,
    data: T,
    config: Record<string, unknown> = {},
  ): Promise<AxiosResponse<T>> {
    return new Promise((resolve, reject) => {
      this.service.post(url, data, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * get 方法
   * @param url
   * @param params
   * @param config
   */
  public get<T>(
    url: string,
    params: Record<string, unknown> = {},
    config: Record<string, unknown> = {},
  ): Promise<AxiosResponse<T>> {
    return new Promise((resolve, reject) => {
      const query = qs.stringify(params);
      this.service.get(`${url}${query ? `?${query}` : ''}`, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * put 方法
   * @param url
   * @param data
   * @param config
   */
  public put<T>(
    url: string,
    data: T,
    config: Record<string, unknown> = {},
  ): Promise<AxiosResponse<T>> {
    return new Promise((resolve, reject) => {
      this.service.put(url, data, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * patch 方法
   * @param url
   * @param data
   * @param config
   */
  public patch<T>(
    url: string,
    data: Partial<T>,
    config: Record<string, unknown> = {},
  ): Promise<AxiosResponse<T>> {
    return new Promise((resolve, reject) => {
      this.service.patch(url, data, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * delete 方法
   * @param url
   */
  public delete(url: string): Promise<AxiosResponse> {
    return new Promise((resolve, reject) => {
      this.service.delete(url).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * 创建唯一实例（单例模式）
   */
  public static getInstance(): MyRequest {
    // 如果 instance 是一个实例 直接返回，如果不是 实例化后返回
    if (!this.resInstance) {
      this.resInstance = new MyRequest();
    }
    return this.resInstance;
  }
}

export default MyRequest.getInstance();
