import axios from 'axios';
import qs from 'qs';

import {
  API_URL,
  requestTimeout,
} from '../../config/net.config';


class MyRequest {
  static resInstance;

  constructor() {
    this.pending = [];
    this.CancelToken = axios.CancelToken;
    this.axiosRequestConfig = {};

    this.requestConfig();
    this.service = axios.create(this.axiosRequestConfig);
    this.interceptorsRequest();
    this.interceptorsResponse();
  }

  /**
   * 初始化配置
   */
  requestConfig() {
    this.axiosRequestConfig = {
      baseURL: API_URL,
      timeout: requestTimeout,
    };
  }

  /**
   * 取消所有还未获取到数据的请求
   * @public
   */
  cancelPreRequests() {
    this.pending.forEach((s) => {
      s.cancel('cancel from client');
    });
    this.pending = [];
  }

  /**
   * 获取请求配置拼装的key
   * @param config
   */
  // eslint-disable-next-line class-methods-use-this
  getKeyOfRequest(config) {
    let key = config.url;
    if (config.params) key += JSON.stringify(config.params);
    if (config.data) key += JSON.stringify(config.data);
    key += `&request_type=${config.method}`;
    return key;
  }

  /**
   * 请求拦截
   */
  interceptorsRequest() {
    this.service.interceptors.request.use(
      (config) => {
        const keyOfRequest = this.getKeyOfRequest(config);
        config.cancelToken = new this.CancelToken((c) => {
          this.pending.push({
            url: keyOfRequest,
            cancel: c,
          });
        });
        return config;
      },
      (error) => Promise.reject(error),
    );
  }

  /**
   * 响应拦截
   */
  interceptorsResponse() {
    this.service.interceptors.response.use(
      (response) => this.handleResponse(response),
      (error) => {
        const { response } = error;
        if (response === undefined) {
          return Promise.reject(new Error(error));
        }
        return this.handleResponse(response);
      },
    );
  }

  /**
   * 请求特殊处理
   * @param response
   */
  handleResponse(response) {
    const code = response.status;
    switch (code) {
      case 200:
        return Promise.resolve(response);
      case 401:
        // TODO token失效,跳转登录页
        break;
      case 403:
        // TODO 没有权限,跳转403页面
        break;
      default:
        break;
    }
    return Promise.reject(response);
  }

  /**
   * post 方法
   * @param url
   * @param data
   * @param config
   */
  post(url, data = {}, config = {}) {
    return new Promise((resolve, reject) => {
      this.service.post(url, data, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * get 方法
   * @param url
   * @param params
   * @param config
   */
  get(url, params = {}, config = {}) {
    return new Promise((resolve, reject) => {
      const query = qs.stringify(params);
      this.service.get(`${url}${query ? `?${query}` : ''}`, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * put 方法
   * @param url
   * @param data
   * @param config
   */
  put(url, data = {}, config = {}) {
    return new Promise((resolve, reject) => {
      this.service.put(url, data, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * patch 方法
   * @param url
   * @param data
   * @param config
   */
  patch(url, data = {}, config = {}) {
    return new Promise((resolve, reject) => {
      this.service.patch(url, data, config).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * delete 方法
   * @param url
   */
  delete(url) {
    return new Promise((resolve, reject) => {
      this.service.delete(url).then((result) => {
        resolve(result);
      }, reject);
    });
  }

  /**
   * 创建唯一实例（单例模式）
   */
  static getInstance() {
    // 如果 instance 是一个实例 直接返回，如果不是 实例化后返回
    if (!this.resInstance) {
      this.resInstance = new MyRequest();
    }
    return this.resInstance;
  }
}

export default MyRequest.getInstance();
